import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomCLAHE(VectorizedBaseRandomLayer):
    """Randomly applies Contrast Limited Adaptive Histogram Equalization to the
    input images.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        factor (int|Sequence[int]|keras_aug.FactorSampler): The range of the
            threshold values for contrast limiting. If the factor is a single
            float value, the range will be ``(1, clip_limit)``. Defaults to
            ``(4, 4)``.
        tile_grid_size (Sequence[int]): The size of grid for histogram
            equalization. Defaults to ``(8, 8)``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `isears/tf_clahe <https://github.com/isears/tf_clahe>`_
    """

    def __init__(
        self,
        value_range,
        factor=(4, 4),
        tile_grid_size=(8, 8),
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.value_range = value_range
        if isinstance(factor, (int, float)):
            factor = (1, factor)
        self.factor = augmentation_utils.parse_factor(
            factor, min_value=1, max_value=None, seed=seed
        )
        self.tile_grid_size = tuple(tile_grid_size)
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        clip_limits = self.factor(shape=(batch_size, 1), dtype=tf.float32)
        return clip_limits

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images, self.value_range, (0, 255), dtype=self.compute_dtype
        )
        inputs_for_clahe_single_image = {
            "images": images,
            "clip_limits": transformations,
        }
        images = tf.map_fn(
            self.clahe_single_image,
            inputs_for_clahe_single_image,
            fn_output_signature=self.compute_dtype,
        )
        images = image_utils.transform_value_range(
            images, (0, 255), self.value_range, dtype=self.compute_dtype
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def compute_clipped_hists(self, clip_limit, tile_shape, hists):
        clip_limit_actual = clip_limit * tf.cast(
            ((tile_shape[0] * tile_shape[1]) / 256), tf.float32
        )
        clip_limit_actual = tf.cast(clip_limit_actual, tf.int32)
        clipped_hists = tf.clip_by_value(
            hists, clip_value_min=0, clip_value_max=clip_limit_actual
        )
        clipped_px_count = tf.math.reduce_sum(hists - clipped_hists, axis=0)
        clipped_hists = tf.cast(clipped_hists, tf.float32)
        clipped_px_count = tf.cast(clipped_px_count, tf.float32)
        clipped_hists = clipped_hists + tf.math.truediv(clipped_px_count, 256)
        return clipped_hists

    def clahe_single_image(self, inputs):
        image = inputs.get("images", None)
        clip_limit = inputs.get("clip_limits", None)

        original_2d_shape = (tf.shape(image)[0], tf.shape(image)[1])
        original_dtype = image.dtype

        # Need image in int32 format for later gather_nd ops
        image = tf.cast(image, tf.int32)

        tile_shape = tf.truediv(original_2d_shape, self.tile_grid_size)
        tile_shape = tf.cast(tf.math.ceil(tile_shape), tf.int32)

        # Reflection-pad image
        pad_y = 0
        pad_x = 0
        pad_y = tf.cond(
            tf.math.equal(original_2d_shape[0] % tile_shape[0], 0),
            lambda: 0,
            lambda: tile_shape[0] - (original_2d_shape[0] % tile_shape[0]),
        )
        pad_x = tf.cond(
            tf.math.equal(original_2d_shape[1] % tile_shape[1], 0),
            lambda: 0,
            lambda: tile_shape[1] - (original_2d_shape[1] % tile_shape[1]),
        )

        image_padded = tf.pad(
            image, [[0, pad_y], [0, pad_x], [0, 0]], "REFLECT"
        )
        all_tiles = tf.space_to_batch(
            input=tf.expand_dims(image_padded, axis=0),
            block_shape=tile_shape,
            paddings=[[0, 0], [0, 0]],
        )

        # Compute per-tile histogram
        # Notes: following operations save some memory compared to original
        # implementation
        ori_tiles_shape = tf.shape(all_tiles)
        all_tiles = tf.reshape(all_tiles, shape=[ori_tiles_shape[0], -1])
        all_tiles = tf.transpose(all_tiles, [1, 0])
        hists = tf.math.bincount(
            all_tiles, minlength=256, maxlength=256, axis=-1
        )
        hists = tf.transpose(hists, [1, 0])
        hists = tf.reshape(
            hists,
            shape=[
                256,
                ori_tiles_shape[1],
                ori_tiles_shape[2],
                ori_tiles_shape[3],
            ],
        )

        clipped_hists = tf.cond(
            tf.math.greater(clip_limit, 0),
            lambda: self.compute_clipped_hists(clip_limit, tile_shape, hists),
            lambda: tf.cast(hists, tf.float32),
        )

        cdf = tf.math.cumsum(clipped_hists, axis=0)
        cdf_min = tf.math.reduce_min(cdf, axis=0)
        numerator = cdf - cdf_min
        denominator = (
            tf.cast(tile_shape[0] * tile_shape[1], tf.float32) - cdf_min
        )
        cdf_normalized = tf.round(
            tf.math.divide_no_nan(numerator, denominator) * (255)
        )
        cdf_normalized = tf.cast(cdf_normalized, tf.int32)

        # Reflection-pad the cdf functions so that we don't have to explicitly
        # deal with corners/edges
        cdf_padded = tf.pad(
            cdf_normalized, [[0, 0], [1, 1], [1, 1], [0, 0]], mode="SYMMETRIC"
        )
        coords = tf.stack(
            tf.meshgrid(
                tf.range(tf.shape(image_padded)[0]),
                tf.range(tf.shape(image_padded)[1]),
                tf.range(tf.shape(image_padded)[2]),
                indexing="ij",
            )
        )

        y_coords = coords[0, :, :]
        x_coords = coords[1, :, :]
        z_coords = coords[2, :, :]

        half_tile_shape = tf.math.floordiv(tile_shape, 2)

        nw_y_component = tf.math.floordiv(
            y_coords - half_tile_shape[0], tile_shape[0]
        )
        nw_x_component = tf.math.floordiv(
            x_coords - half_tile_shape[1], tile_shape[1]
        )

        # Need to correct negative values because negative-indexing for
        # gather_nd ops not supported on all processors
        # (cdf is padded to account for this)
        nw_y_component = nw_y_component + 1
        nw_x_component = nw_x_component + 1
        ne_y_component = nw_y_component
        ne_x_component = nw_x_component + 1
        sw_y_component = nw_y_component + 1
        sw_x_component = nw_x_component
        se_y_component = sw_y_component
        se_x_component = sw_x_component + 1

        def cdf_transform(x_comp, y_comp):
            gatherable = tf.stack(
                [image_padded, y_comp, x_comp, z_coords], axis=-1
            )
            return tf.cast(tf.gather_nd(cdf_padded, gatherable), tf.float32)

        nw_transformed = cdf_transform(nw_x_component, nw_y_component)
        ne_transformed = cdf_transform(ne_x_component, ne_y_component)
        sw_transformed = cdf_transform(sw_x_component, sw_y_component)
        se_transformed = cdf_transform(se_x_component, se_y_component)

        a = (y_coords - half_tile_shape[0]) % tile_shape[0]
        a = tf.cast(tf.math.truediv(a, tile_shape[0]), tf.float32)
        b = (x_coords - half_tile_shape[1]) % tile_shape[1]
        b = tf.cast(tf.math.truediv(b, tile_shape[1]), tf.float32)

        # Interpolate
        interpolated = (a * (b * se_transformed + (1 - b) * sw_transformed)) + (
            1 - a
        ) * (b * ne_transformed + (1 - b) * nw_transformed)

        # Return image to original size and dtype
        interpolated = interpolated[
            0 : original_2d_shape[0], 0 : original_2d_shape[1], :
        ]
        interpolated = tf.cast(tf.round(interpolated), original_dtype)
        return interpolated

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "factor": self.factor,
                "tile_grid_size": self.tile_grid_size,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class PadIfNeeded(VectorizedBaseRandomLayer):
    """Pads the images if needed.

    PadIfNeeded can be configured by specifying the height/width or the
    divisors. The images will be padded to ``(min_height, min_width)`` or the
    size of the both sides to be divisible by ``height_divisor`` and
    ``width_divisor``. PadIfNeeded is required to specify
    ``min_height`` or ``height_divisor`` and ``min_width`` or
    ``width_divisor``, respectively.

    Args:
        min_height (int, optional): The height of result image.
        min_width (int, optional): The width of result image.
        height_divisor (int, optional): The divisor that ensures image height is
            divisible by.
        width_divisor (int, optional): The divisor that ensures image width is
            divisible by.
        position (str, optional): The padding method.
            Supported values: ``"center", "top_left", "top_right", "bottom_left", "bottom_right", "random"``.
            Defaults to ``"center"``.
        padding_value (int|float, optional): The padding value.
            Defaults to ``0``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `Albumentations <https://github.com/albumentations-team/albumentations>`_
    """  # noqa: E501

    def __init__(
        self,
        min_height=None,
        min_width=None,
        pad_height_divisor=None,
        pad_width_divisor=None,
        position="center",
        padding_value=0,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if (min_height is None or min_width is None) and (
            pad_height_divisor is None or pad_width_divisor is None
        ):
            raise ValueError(
                "`PadIfNeeded()` expects at most one of "
                "(`min_height`, `min_width`) or "
                "(`pad_height_divisor`, `pad_width_divisor`) to be set."
            )

        self.min_height = min_height
        self.min_width = min_width
        self.pad_height_divisor = pad_height_divisor
        self.pad_width_divisor = pad_width_divisor
        self.position = augmentation_utils.get_padding_position(position)
        self.padding_value = padding_value
        self.bounding_box_format = bounding_box_format
        self.seed = seed

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        heights, widths = augmentation_utils.get_images_shape(images)

        if self.min_height is not None:
            tops = tf.where(
                heights < self.min_height,
                tf.cast((self.min_height - heights) / 2, heights.dtype),
                tf.zeros_like(heights, dtype=heights.dtype),
            )
            bottoms = tf.where(
                heights < self.min_height,
                self.min_height - heights - tops,
                tf.zeros_like(heights, dtype=heights.dtype),
            )
        else:
            pad_remaineds = heights % self.pad_height_divisor
            pad_rows = tf.where(
                tf.math.greater(pad_remaineds, 0),
                self.pad_height_divisor - pad_remaineds,
                tf.zeros_like(pad_remaineds, dtype=pad_remaineds.dtype),
            )
            pad_rows = tf.cast(pad_rows, dtype=tf.float32)
            tops = tf.round(pad_rows / 2.0)
            bottoms = tf.cast(pad_rows - tops, dtype=tf.int32)
            tops = tf.cast(tops, dtype=tf.int32)

        if self.min_width is not None:
            lefts = tf.where(
                widths < self.min_width,
                tf.cast((self.min_width - widths) / 2, widths.dtype),
                tf.zeros_like(widths, dtype=widths.dtype),
            )
            rights = tf.where(
                widths < self.min_width,
                self.min_width - widths - lefts,
                tf.zeros_like(widths, dtype=widths.dtype),
            )
        else:
            pad_remaineds = widths % self.pad_width_divisor
            pad_cols = tf.where(
                tf.math.greater(pad_remaineds, 0),
                self.pad_width_divisor - pad_remaineds,
                tf.zeros_like(pad_remaineds, dtype=pad_remaineds.dtype),
            )
            pad_cols = tf.cast(pad_cols, dtype=tf.float32)
            lefts = tf.round(pad_cols / 2.0)
            rights = tf.cast(pad_cols - lefts, dtype=tf.int32)
            lefts = tf.cast(lefts, dtype=tf.int32)

        (tops, bottoms, lefts, rights) = augmentation_utils.get_position_params(
            tops, bottoms, lefts, rights, self.position, self._random_generator
        )

        return {
            "pad_tops": tops,
            "pad_bottoms": bottoms,
            "pad_lefts": lefts,
            "pad_rights": rights,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        pad_top = transformations["pad_tops"][0][0]
        pad_bottom = transformations["pad_bottoms"][0][0]
        pad_left = transformations["pad_lefts"][0][0]
        pad_right = transformations["pad_rights"][0][0]
        paddings = tf.stack(
            (
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
                tf.stack((pad_top, pad_bottom)),
                tf.stack((pad_left, pad_right)),
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
            )
        )
        images = tf.pad(
            images, paddings=paddings, constant_values=self.padding_value
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`PadIfNeeded()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`PadIfNeeded(..., bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
        )

        pad_tops = tf.cast(transformations["pad_tops"], dtype=tf.float32)
        pad_lefts = tf.cast(transformations["pad_lefts"], dtype=tf.float32)
        x1s, y1s, x2s, y2s = tf.split(bounding_boxes["boxes"], 4, axis=-1)
        x1s += tf.expand_dims(pad_lefts, axis=1)
        y1s += tf.expand_dims(pad_tops, axis=1)
        x2s += tf.expand_dims(pad_lefts, axis=1)
        y2s += tf.expand_dims(pad_tops, axis=1)
        outputs = tf.concat([x1s, y1s, x2s, y2s], axis=-1)

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = outputs
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=images,
        )
        return bounding_boxes

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        pad_top = transformations["pad_tops"][0][0]
        pad_bottom = transformations["pad_bottoms"][0][0]
        pad_left = transformations["pad_lefts"][0][0]
        pad_right = transformations["pad_rights"][0][0]
        paddings = tf.stack(
            (
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
                tf.stack((pad_top, pad_bottom)),
                tf.stack((pad_left, pad_right)),
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
            )
        )
        segmentation_masks = tf.pad(
            segmentation_masks, paddings=paddings, constant_values=0
        )
        return segmentation_masks

    def augment_segmentation_mask_single(self, inputs):
        segmentation_mask = inputs.get(
            augmentation_utils.SEGMENTATION_MASKS, None
        )
        transformation = inputs.get("transformation", None)
        pad_top = transformation["pad_tops"][0]
        pad_bottom = transformation["pad_bottoms"][0]
        pad_left = transformation["pad_lefts"][0]
        pad_right = transformation["pad_rights"][0]
        paddings = tf.stack(
            (
                tf.stack((pad_top, pad_bottom)),
                tf.stack((pad_left, pad_right)),
                tf.zeros(shape=(2,), dtype=pad_top.dtype),
            )
        )
        segmentation_mask = tf.pad(
            segmentation_mask, paddings=paddings, constant_values=0
        )
        return segmentation_mask

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "min_height": self.min_height,
                "min_width": self.min_width,
                "pad_height_divisor": self.pad_height_divisor,
                "pad_width_divisor": self.pad_width_divisor,
                "position": self.position,
                "padding_value": self.padding_value,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



import typing

import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomResize(VectorizedBaseRandomLayer):
    """Randomly resizes the images in a batch manner.

    This layer is useful for multi-scale training.

    Notes:
        The aspect ratio might be different from the original images.

    Args:
        heights (list(int)): The heights to be sampled for the result image.
        widths (list(int), optional): The widths to be sampled for the result
            image. Defaults to ``None``. If setting ``None``, ``widths`` will
            be the same as ``heights``.
        interpolation (str, optional): The interpolation mode.
            Supported values: ``"nearest", "bilinear"``. Defaults to
            ``"bilinear"``.
        antialias (bool, optional): Whether to use antialias. Defaults to
            ``False``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.
    """  # noqa: E501

    def __init__(
        self,
        heights,
        widths=None,
        interpolation="bilinear",
        antialias=False,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.heights = heights
        if widths is None:
            self.widths = heights
        else:
            self.widths = widths
        if not isinstance(self.heights, typing.Sequence) or not isinstance(
            self.widths, typing.Sequence
        ):
            raise ValueError(
                "RandomResize expects `heights` and `widths` to be a list of "
                f"int. Received: `heights`={heights}, `widths`={widths}"
            )
        if len(self.heights) != len(self.widths):
            raise ValueError(
                "RandomResize expects `heights` and `widths` to be same "
                f"length. Received: `heights`={heights}, `widths`={widths}"
            )
        self.interpolation = augmentation_utils.get_interpolation(interpolation)
        self.antialias = antialias
        self.bounding_box_format = bounding_box_format
        self.seed = seed

        self._heights = tf.convert_to_tensor(self.heights, dtype=tf.int32)
        self._widths = tf.convert_to_tensor(self.widths, dtype=tf.int32)

        # set force_output_dense_images=True because the output images must
        # have same shape (B, height, width, C)
        self.force_output_dense_images = True

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        # sample 1 height and width for the batch
        indice = self._random_generator.random_uniform(
            shape=(1,),
            minval=0,
            maxval=len(self.heights),
            dtype=tf.int32,
        )
        height = tf.gather(self._heights, indice)
        width = tf.gather(self._widths, indice)
        scaled_sizes = tf.stack([height, width], axis=-1)
        scaled_sizes = tf.tile(scaled_sizes, multiples=(batch_size, 1))
        return {"scaled_sizes": scaled_sizes}

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        scaled_sizes = transformations["scaled_sizes"]

        # currently, only support same size within dense batch
        # this layer might produce varying size with ragged batch
        new_height = scaled_sizes[0][0]
        new_width = scaled_sizes[0][1]
        images = tf.image.resize(
            images,
            size=(new_height, new_width),
            method=self.interpolation,
            antialias=self.antialias,
        )
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`Resize()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`Resize(..., bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)

        # resize
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="rel_xyxy",
            images=raw_images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="rel_xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=images,
        )
        return bounding_boxes

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        scaled_size = transformations["scaled_sizes"]
        new_height = scaled_size[0][0]
        new_width = scaled_size[0][1]
        # resize
        segmentation_masks = tf.image.resize(
            segmentation_masks,
            size=(new_height, new_width),
            method="nearest",
        )
        return tf.cast(segmentation_masks, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "heights": self.heights,
                "widths": self.widths,
                "interpolation": self.interpolation,
                "antialias": self.antialias,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_aug import layers


class EqualizeTest(tf.test.TestCase, parameterized.TestCase):
    def test_return_shapes_inside_model(self):
        layer = layers.Equalize(value_range=(0, 255))
        inp = keras.layers.Input(shape=[4, 4, 5])
        out = layer(inp)
        model = keras.models.Model(inp, out)

        self.assertEqual(model.layers[-1].output_shape, (None, 4, 4, 5))

    def test_equalizes_to_all_bins(self):
        xs = tf.random.uniform((2, 32, 32, 3), 0, 255, seed=2023)
        layer = layers.Equalize(value_range=(0, 255))
        xs = layer(xs)

        for i in range(0, 256):
            self.assertTrue(tf.math.reduce_any(xs == i))

    @parameterized.named_parameters(
        ("float32", tf.float32), ("int32", tf.int32), ("int64", tf.int64)
    )
    def test_input_dtypes(self, dtype):
        xs = tf.random.uniform((2, 32, 32, 3), 0, 255, dtype=dtype, seed=2023)
        layer = layers.Equalize(value_range=(0, 255))
        xs = layer(xs)

        for i in range(0, 256):
            self.assertTrue(tf.math.reduce_any(xs == i))
        self.assertAllInRange(xs, 0, 255)

    @parameterized.named_parameters(("0_255", 0, 255), ("0_1", 0, 1))
    def test_output_range(self, lower, upper):
        xs = tf.random.uniform((2, 4, 4, 3), lower, upper, dtype=tf.float32)
        layer = layers.Equalize(value_range=(lower, upper))
        xs = layer(xs)
        self.assertAllInRange(xs, lower, upper)



import tensorflow as tf

from keras_aug import layers


class TrivialAugmentWideTest(tf.test.TestCase):
    def test_runs(self):
        trivial_augment = layers.TrivialAugmentWide(
            value_range=(0, 255),
            seed=2023,
        )
        xs = tf.ones((2, 4, 4, 3))

        ys = trivial_augment(xs)

        self.assertEqual(ys.shape, (2, 4, 4, 3))



"""
Copied from:
https://github.com/keras-team/keras/blob/4829ddfc5fdcb84efe19b5e571f4be8dda89a3e1/keras/engine/base_layer.py

Remove __init__ docstring to prevent docstring population
"""  # noqa: E501

import tensorflow as tf
from tensorflow import keras

from keras_aug.core.keras_random_generator import KerasRandomGenerator


class BaseRandomLayer(keras.layers.Layer):
    """A layer handle the random number creation and savemodel behavior.

    Args:
          seed: optional integer, used to create RandomGenerator.
          force_generator: boolean, default to False, whether to force the
            RandomGenerator to use the code branch of tf.random.Generator.
          rng_type: string, the rng type that will be passed to backend
            RandomGenerator. Default to `None`, which will allow RandomGenerator
            to choose types by itself. Valid values are "stateful", "stateless",
            "legacy_stateful".
          **kwargs: other keyword arguments that will be passed to the parent
            *class
    """

    @tf.__internal__.tracking.no_automatic_dependency_tracking
    def __init__(
        self, seed=None, force_generator=False, rng_type=None, **kwargs
    ):
        super().__init__(**kwargs)
        self._random_generator = KerasRandomGenerator(
            seed, force_generator=force_generator, rng_type=rng_type
        )

    def build(self, input_shape):
        super().build(input_shape)
        self._random_generator._maybe_init()

    def _trackable_children(self, save_type="checkpoint", **kwargs):
        if save_type == "savedmodel":
            cache = kwargs["cache"]
            # TODO(b/213628533): This must be called before super() to ensure
            # that any input shape changes are applied before getting the config
            # of the model.
            children = self._trackable_saved_model_saver.trackable_children(
                cache
            )
            # This method exposes the self._random_generator to SavedModel only
            # (not layer.weights and checkpoint).
            children["_random_generator"] = self._random_generator
        else:
            children = {}
        children.update(super()._trackable_children(save_type, **kwargs))
        return children

    def _lookup_dependency(self, name):
        # When loading from a Keras SavedModel load, make sure that the loader
        # can find the random generator, otherwise the loader will assume that
        # it does not exist, and will try to create a new generator.
        if name == "_random_generator":
            return self._random_generator
        else:
            return super()._lookup_dependency(name)



import numpy as np
import tensorflow as tf

from keras_aug import layers
from keras_aug.datapoints import bounding_box


class RandomRotationTest(tf.test.TestCase):
    regular_args = {
        "factor": 10,
        "fill_mode": "constant",
        "fill_value": 0,
        "interpolation": "bilinear",
        "bounding_box_format": "xyxy",
    }
    no_aug_args = {
        "factor": 0.0,
        "fill_mode": "constant",
        "fill_value": 0,
        "interpolation": "bilinear",
        "bounding_box_format": "xyxy",
    }

    def test_no_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = layers.RandomRotate(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.RandomRotate(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, rtol=1e-5, atol=1e-5)

        layer = layers.RandomRotate(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_rotate_unbatched_image(self):
        input_image = np.reshape(np.arange(0, 25), (5, 5, 1)).astype(np.float32)
        # 180 rotation
        args = self.no_aug_args.copy()
        args.update({"factor": (180, 180)})
        layer = layers.RandomRotate(**args)
        expected_output = np.asarray(
            [
                [24, 23, 22, 21, 20],
                [19, 18, 17, 16, 15],
                [14, 13, 12, 11, 10],
                [9, 8, 7, 6, 5],
                [4, 3, 2, 1, 0],
            ]
        ).astype(np.float32)

        output_image = layer(input_image)
        expected_output = np.reshape(expected_output, (5, 5, 1))

        self.assertAllClose(expected_output, output_image)

    def test_rotate_bounding_boxes(self):
        input_image = np.random.random((512, 512, 3)).astype(np.float32)
        bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[200, 200, 400, 400], [100, 100, 300, 300]], dtype=tf.float32
            ),
            "classes": tf.convert_to_tensor([1, 2], dtype=tf.float32),
        }
        input = {"images": input_image, "bounding_boxes": bounding_boxes}
        expected_bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[112.0, 112.0, 312.0, 312.0], [212.0, 212.0, 412.0, 412.0]],
                dtype=tf.float32,
            ),
            "classes": tf.convert_to_tensor([1, 2], dtype=tf.float32),
        }
        # 180 rotation
        args = self.no_aug_args.copy()
        args.update({"factor": (180, 180)})
        layer = layers.RandomRotate(**args)

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(expected_bounding_boxes, output["bounding_boxes"])

    def test_rotate_ragged_bounding_boxes(self):
        input_image = np.random.random((2, 512, 512, 3)).astype(np.float32)
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[200, 200, 400, 400], [100, 100, 300, 300]],
                    [[200, 200, 400, 400]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant(
                [[0, 0], [0]],
                dtype=tf.float32,
            ),
        }
        input = {"images": input_image, "bounding_boxes": bounding_boxes}
        expected_output = {
            "boxes": tf.ragged.constant(
                [
                    [
                        [112.0, 112.0, 312.0, 312.0],
                        [212.0, 212.0, 412.0, 412.0],
                    ],
                    [[112.0, 112.0, 312.0, 312.0]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant(
                [[0, 0], [0]],
                dtype=tf.float32,
            ),
        }
        # 180 rotation
        args = self.no_aug_args.copy()
        args.update({"factor": (180, 180)})
        layer = layers.RandomRotate(**args)

        output = layer(input)

        expected_output = bounding_box.to_dense(expected_output)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"],
            output["bounding_boxes"]["classes"],
        )

    def test_rotate_sparse_segmentation_mask(self):
        num_classes = 8

        input_images = np.random.random((2, 20, 20, 3)).astype(np.float32)
        # Masks are all 0s or 8s, to verify that when we rotate we don't do bad
        # mask interpolation to either a 0 or a 7
        masks = np.random.randint(2, size=(2, 20, 20, 1)) * (num_classes - 1)
        inputs = {"images": input_images, "segmentation_masks": masks}

        # 90 rotation
        args = self.no_aug_args.copy()
        args.update({"factor": (90, 90)})
        layer = layers.RandomRotate(**args)
        outputs = layer(inputs)
        expected_masks = np.rot90(masks, axes=(1, 2))
        self.assertAllClose(expected_masks, outputs["segmentation_masks"])

        # 45-degree rotation. Only verifies that no interpolation takes place.
        # 90 rotation
        args = self.no_aug_args.copy()
        args.update({"factor": (0.125, 0.125)})
        outputs = layer(inputs)
        self.assertAllInSet(outputs["segmentation_masks"], [0, 7])

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        args = self.no_aug_args.copy()
        args.update({"factor": (90, 90)})
        layer = layers.RandomRotate(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        args = self.no_aug_args.copy()
        args.update({"factor": (90, 90)})
        layer = layers.RandomRotate(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(
            isinstance(result["segmentation_masks"], tf.RaggedTensor)
        )
        self.assertAllInSet(
            result["segmentation_masks"].to_tensor(), tf.range(0, 10)
        )



from keras_aug.core.factor_sampler.constant_factor_sampler import (
    ConstantFactorSampler,
)
from keras_aug.core.factor_sampler.factor_sampler import FactorSampler
from keras_aug.core.factor_sampler.normal_factor_sampler import (
    NormalFactorSampler,
)
from keras_aug.core.factor_sampler.signed_constant_factor_sampler import (
    SignedConstantFactorSampler,
)
from keras_aug.core.factor_sampler.signed_normal_factor_sampler import (
    SignedNormalFactorSampler,
)
from keras_aug.core.factor_sampler.uniform_factor_sampler import (
    UniformFactorSampler,
)



"""
Most of these codes come from KerasCV.
"""
import tensorflow as tf

from keras_aug.datapoints import bounding_box


class ValidateTest(tf.test.TestCase):
    def test_raises_nondict(self):
        with self.assertRaisesRegex(
            ValueError, "Expected `bounding_boxes` to be a dictionary, got "
        ):
            bounding_box.validate_format(tf.ones((4, 3, 6)))

    def test_mismatch_dimensions(self):
        with self.assertRaisesRegex(
            ValueError,
            "Expected `boxes` and `classes` to have matching dimensions",
        ):
            bounding_box.validate_format(
                {"boxes": tf.ones((4, 3, 6)), "classes": tf.ones((4, 6))}
            )

    def test_bad_keys(self):
        with self.assertRaisesRegex(ValueError, "containing keys"):
            bounding_box.validate_format(
                {
                    "box": [1, 2, 3],
                    "class": [1234],
                }
            )



import tensorflow as tf

from keras_aug import layers


class RepeatedAugmentationTest(tf.test.TestCase):
    def test_output_shapes(self):
        repeated_augment = layers.RepeatedAugment(
            layers=[
                layers.RandAugment(value_range=(0, 255)),
                layers.RandomFlip(),
            ]
        )
        inputs = {
            "images": tf.ones((4, 8, 8, 3)),
            "labels": tf.ones((4,)),
        }
        outputs = repeated_augment(inputs)

        self.assertEqual(outputs["images"].shape, (8, 8, 8, 3))
        self.assertEqual(outputs["labels"].shape, (8,))

    def test_with_mix_up(self):
        repeated_augment = layers.RepeatedAugment(
            layers=[
                layers.RandAugment(value_range=(0, 255)),
                layers.MixUp(),
            ]
        )
        inputs = {
            "images": tf.ones((4, 8, 8, 3)),
            "labels": tf.ones((4, 10)),
        }
        outputs = repeated_augment(inputs)

        self.assertEqual(outputs["images"].shape, (8, 8, 8, 3))
        self.assertEqual(outputs["labels"].shape, (8, 10))



import tensorflow as tf

import keras_aug


class ConstantFactorSamplerTest(tf.test.TestCase):
    def test_sample(self):
        factor = keras_aug.ConstantFactorSampler(0.3)

        self.assertEqual(factor(), 0.3)

    def test_config(self):
        factor = keras_aug.ConstantFactorSampler(0.3)

        config = factor.get_config()

        self.assertEqual(config["value"], 0.3)

    def test_from_config(self):
        factor = keras_aug.ConstantFactorSampler(0.3)
        config = factor.get_config()

        factor2 = keras_aug.ConstantFactorSampler.from_config(config)
        config = factor2.get_config()

        self.assertEqual(config["value"], 0.3)



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomChannelDropout(VectorizedBaseRandomLayer):
    """Randomly drop channels of the input images.

    Args:
        factor (float|Sequence[float]|keras_aug.FactorSampler): The range from
            which we choose the number of channels to drop. Defaults to
            ``(0, 2)``.
        fill_value (int|float, optional): The value to be filled for dropped
            channel. Defaults to ``0``.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `Albumentations <https://github.com/albumentations-team/albumentations>`_
    """  # noqa: E501

    def __init__(self, factor=(0, 2), fill_value=0, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        if isinstance(factor, (tuple, list)):
            lower = factor[0]
            upper = factor[1] + 1
        else:
            lower = 0
            upper = factor + 1
        self.factor_input = factor
        self.factor = augmentation_utils.parse_factor(
            (lower, upper), min_value=0, max_value=None, seed=seed
        )
        self.fill_value = fill_value
        self.seed = seed

        self.upper = upper

    def get_random_transformation_batch(self, batch_size, **kwargs):
        channels_to_drop = self.factor(shape=(batch_size, 1), dtype=tf.int32)
        return channels_to_drop

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = tf.cast(images, dtype=self.compute_dtype)
        channel = images.shape[-1]
        indices = transformations

        if self.upper > channel:
            raise ValueError(
                f"The number of the channel: {channel} must <= the upper "
                f"bound: {self.upper}."
            )

        drop_matrix = tf.one_hot(
            indices,
            channel,
            on_value=0,
            off_value=1,
            axis=-1,
            dtype=self.compute_dtype,
        )
        drop_matrix = tf.expand_dims(drop_matrix, axis=1)
        fill_value_matrix = (
            tf.one_hot(
                indices,
                channel,
                on_value=1,
                off_value=0,
                axis=-1,
                dtype=self.compute_dtype,
            )
            * self.fill_value
        )
        fill_value_matrix = tf.expand_dims(fill_value_matrix, axis=1)
        return images * drop_matrix + fill_value_matrix

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = {
            "factor": self.factor_input,
            "fill_value": self.fill_value,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomBlur(VectorizedBaseRandomLayer):
    """Randomly blurs the images using random-sized kernels.

    This layer applies a mean filter with varying kernel sizes to blur the
    images. The sampled kernel sizes are always odd numbers.

    Args:
        factor (int|Sequence[int]|keras_aug.FactorSampler): The kernel size range
            for blurring the input image. If the factor is a single value, the
            range will be ``(1, factor)``. The value range of the factor should
            be in ``(1, +inf)``. When sampled kernel size=``1``, there is no
            blur effect.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `Albumentations <https://github.com/albumentations-team/albumentations>`_
    """  # noqa: E501

    def __init__(
        self,
        factor,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if isinstance(factor, (tuple, list)):
            factor_range = (0, (factor[1] - factor[0]) // 2 + 1)
            factor_bias = factor[0]
        else:
            factor_range = (0, (factor - 1) // 2 + 1)
            factor_bias = 1
        if factor_range[1] < 0 or factor_bias < 1:
            raise ValueError(
                "RandomBlur expects `factor` to be in range `(1, inf)`. Got: "
                f"`factor` = {factor}"
            )
        self.factor_input = factor

        self.factor_bias = factor_bias
        self.factor = augmentation_utils.parse_factor(
            factor_range,
            min_value=0,
            max_value=None,
            center_value=None,
            seed=seed,
        )
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        blur_kernel_sizes = self.factor(shape=(batch_size, 1), dtype=tf.int32)
        # [0, k] => [0, ..., 2k+1] ensures only odd numbers
        blur_kernel_sizes = blur_kernel_sizes * 2 + self.factor_bias
        return blur_kernel_sizes

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        inputs_for_blur_single_image = {
            "images": images,
            "blur_kernel_sizes": transformations,
        }
        images = tf.map_fn(
            self.blur_single_image,
            inputs_for_blur_single_image,
            fn_output_signature=tf.float32,
        )
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def blur_single_image(self, inputs):
        image = inputs.get("images", None)
        image = tf.expand_dims(image, axis=0)
        image = tf.cast(image, tf.float32)
        blur_kernel_size = inputs.get("blur_kernel_sizes", None)
        blur_kernel_size = blur_kernel_size[0]

        mean_filter = tf.ones(
            shape=(blur_kernel_size, blur_kernel_size, 1, image.shape[-1])
        ) / tf.math.square(tf.cast(blur_kernel_size, tf.float32))
        blurred_image = tf.nn.conv2d(
            image, mean_filter, strides=1, padding="SAME"
        )
        return tf.squeeze(blurred_image, axis=0)

    def get_config(self):
        config = super().get_config()
        config.update({"factor": self.factor_input, "seed": self.seed})
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomGamma(VectorizedBaseRandomLayer):
    """Randomly adjusts gamma of the input images.

    This layer will randomly increase/reduce the gamma for the input images by
    the equation: ``y = x ** factor``. Gamma is adjusted independently of each
    image. The image is adjusted by converting the pixel value range to
    ``[0, 1]`` and applying RandomGamma. The image is then converted back to the
    original value range.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        factor (float|Sequence[float]|keras_aug.FactorSampler): The range of the
            gamma factor. When represented as a single float, the
            factor will be picked between ``[1.0 - lower, 1.0 + upper]``.
            ``1.0`` will give the original image.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.
    """

    def __init__(
        self,
        value_range,
        factor,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.factor = augmentation_utils.parse_factor(
            factor, max_value=None, center_value=1.0, seed=seed
        )
        self.value_range = value_range
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        factors = self.factor(shape=(batch_size, 1), dtype=self.compute_dtype)
        return factors

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images, self.value_range, (0.0, 1.0), dtype=self.compute_dtype
        )
        factors = transformations
        images = tf.pow(images, factors[:, :, tf.newaxis, tf.newaxis])
        images = image_utils.transform_value_range(
            images, (0.0, 1.0), self.value_range, dtype=self.compute_dtype
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "factor": self.factor,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomColorJitter(VectorizedBaseRandomLayer):
    """Randomly applies brightness, contrast, saturation and hue image
    processing operation sequentially and randomly on the input images. It
    expects input as RGB image.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        brightness_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The
            range of the brightness factor. When represented as a single float,
            the factor will be picked between ``[1.0 - lower, 1.0 + upper]``.
            ``0.0`` will make image be black. ``1.0`` will make image be white.
            Defaults to ``None``.
        contrast_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The range
            of the contrast factor. When represented as a single float, the
            factor will be picked between ``[1.0 - lower, 1.0 + upper]``.
            ``0.0`` gives solid gray image. ``1.0`` gives the original image
            while ``2.0`` increases the contrast by a factor of 2.
            Defaults to ``None``.
        saturation_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The
            range of the saturation factor. When represented as a single float,
            the factor will be picked between ``[1.0 - lower, 1.0 + upper]``.
            ``1.0`` will give the original image. ``0.0`` makes the image to be
            fully grayscale. ``2.0`` will enhance the saturation by a factor of
            2. Defaults to ``None``.
        hue_factor (float|Sequence[float]|keras_aug.FactorSampler, optional): The range of
            the hue factor. When represented as a single float, the factor will
            be picked between ``[0.0 - lower, 0.0 + upper]``. ``0.0`` means no
            shift. ``-0.5`` or ``0.5`` gives an image with complementary colors.
            Defaults to ``None``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `torchvision <https://github.com/pytorch/vision>`_
        - `Tensorflow Model augment <https://github.com/tensorflow/models/blob/master/official/vision/ops/augment.py>`
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        value_range,
        brightness_factor=None,
        contrast_factor=None,
        saturation_factor=None,
        hue_factor=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.brightness_factor = None
        self.contrast_factor = None
        self.saturation_factor = None
        self.hue_factor = None
        if brightness_factor:
            self.brightness_factor = augmentation_utils.parse_factor(
                brightness_factor, max_value=None, center_value=1, seed=seed
            )
        if contrast_factor is not None:
            self.contrast_factor = augmentation_utils.parse_factor(
                contrast_factor, max_value=None, center_value=1, seed=seed
            )
        if saturation_factor is not None:
            self.saturation_factor = augmentation_utils.parse_factor(
                saturation_factor, max_value=None, center_value=1, seed=seed
            )
        if hue_factor is not None:
            self.hue_factor = augmentation_utils.parse_factor(
                hue_factor,
                min_value=-0.5,
                max_value=0.5,
                center_value=0,
                seed=seed,
            )
        self.value_range = value_range
        self.seed = seed

        # decide whether to enable the augmentation
        self._enable_brightness = augmentation_utils.is_factor_working(
            self.brightness_factor, not_working_value=1.0
        )
        self._enable_contrast = augmentation_utils.is_factor_working(
            self.contrast_factor, not_working_value=1.0
        )
        self._enable_saturation = augmentation_utils.is_factor_working(
            self.saturation_factor, not_working_value=1.0
        )
        self._enable_hue = augmentation_utils.is_factor_working(
            self.hue_factor, not_working_value=0.0
        )

    def get_random_transformation_batch(self, batch_size, **kwargs):
        factor_shape = (batch_size, 1, 1, 1)
        # dummy
        brightness_factors = tf.zeros(factor_shape)
        contrast_factors = tf.zeros(factor_shape)
        saturation_factors = tf.zeros(factor_shape)
        hue_factors = tf.zeros(factor_shape)
        if self._enable_brightness:
            brightness_factors = self.brightness_factor(
                shape=factor_shape, dtype=self.compute_dtype
            )
        if self._enable_contrast:
            contrast_factors = self.contrast_factor(
                shape=factor_shape, dtype=self.compute_dtype
            )
        if self._enable_saturation:
            saturation_factors = self.saturation_factor(
                shape=factor_shape, dtype=self.compute_dtype
            )
        if self._enable_hue:
            hue_factors = self.hue_factor(
                shape=factor_shape, dtype=self.compute_dtype
            )
        return {
            "brightness_factors": brightness_factors,
            "contrast_factors": contrast_factors,
            "saturation_factors": saturation_factors,
            "hue_factors": hue_factors,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        transformations = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        images = self.augment_images(
            images=images, transformations=transformations, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images, self.value_range, (0, 255), dtype=self.compute_dtype
        )
        if self._enable_brightness:
            images = self.adjust_brightness(images, transformations)
        if self._enable_contrast:
            images = self.adjust_contrast(images, transformations)
        if self._enable_saturation:
            images = self.adjust_saturation(images, transformations)
        if self._enable_hue:
            images = self.adjust_hue(images, transformations)
        images = image_utils.transform_value_range(
            images, (0, 255), self.value_range, dtype=self.compute_dtype
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def adjust_brightness(self, images, transformations):
        brightness_factors = transformations["brightness_factors"]
        images *= brightness_factors
        images = tf.clip_by_value(images, 0, 255)
        return images

    def adjust_contrast(self, images, transformations):
        # cast to float32 to avoid numerical issue
        contrast_factors = tf.cast(
            transformations["contrast_factors"], dtype=tf.float32
        )
        images = tf.cast(images, dtype=tf.float32)
        degenerates = image_utils.rgb_to_grayscale(images)
        degenerates = tf.reduce_mean(degenerates, axis=(1, 2, 3), keepdims=True)
        images = image_utils.blend(degenerates, images, contrast_factors)
        images = tf.clip_by_value(images, 0, 255)
        return tf.cast(images, dtype=self.compute_dtype)

    def adjust_saturation(self, images, transformations):
        saturation_factors = transformations["saturation_factors"]
        degenerates = image_utils.rgb_to_grayscale(images)
        images = image_utils.blend(
            degenerates, images, saturation_factors, (0, 255)
        )
        images = tf.clip_by_value(images, 0, 255)
        return images

    def adjust_hue(self, images, transformations):
        # cast to float32 to avoid numerical issue
        images = tf.image.rgb_to_hsv(tf.cast(images, dtype=tf.float32))
        hue_factors = tf.cast(transformations["hue_factors"], dtype=tf.float32)
        h_channels = tf.math.floormod(images[..., 0:1] + hue_factors, 1.0)
        images = tf.concat(
            [h_channels, images[..., 1:2], images[..., 2:3]], axis=-1
        )
        images = tf.image.hsv_to_rgb(images)
        images = tf.clip_by_value(images, 0, 255)
        return tf.cast(images, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "brightness_factor": self.brightness_factor,
                "contrast_factor": self.contrast_factor,
                "saturation_factor": self.saturation_factor,
                "hue_factor": self.hue_factor,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class AutoContrast(VectorizedBaseRandomLayer):
    """Performs the AutoContrast operation on the input images.

    Auto contrast stretches the values of an image across the entire available
    ``value_range``. This makes differences between pixels more obvious. An
    example of this is if an image only has values ``[0, 1]`` out of the range
    ``[0, 255]``, auto contrast will change the ``1`` values to be ``255``.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, value_range, **kwargs):
        super().__init__(**kwargs)
        self.value_range = value_range

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        images = self.augment_images(
            images=images, transformations=transformation, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images,
            original_range=self.value_range,
            target_range=(0, 255),
            dtype=self.compute_dtype,
        )
        lows = tf.reduce_min(images, axis=(1, 2), keepdims=True)
        highs = tf.reduce_max(images, axis=(1, 2), keepdims=True)
        scales = 255.0 / (highs - lows)
        eq_idxs = tf.math.is_inf(scales)
        lows = tf.where(eq_idxs, 0.0, lows)
        scales = tf.where(eq_idxs, 1.0, scales)
        images = tf.clip_by_value((images - lows) * scales, 0, 255)
        images = image_utils.transform_value_range(
            images,
            original_range=(0, 255),
            target_range=self.value_range,
            dtype=self.compute_dtype,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update({"value_range": self.value_range})
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class Grayscale(VectorizedBaseRandomLayer):
    """Grayscale transforms RGB images to grayscale images.

    Args:
        output_channels (int, optional): The number of the color channels of
            the outputs. Defaults to ``3``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, output_channels=3, **kwargs):
        super().__init__(**kwargs)
        if output_channels not in (1, 3):
            raise ValueError(
                "Received invalid argument output_channels. "
                f"output_channels must be in 1 or 3. Got {output_channels}"
            )
        self.output_channels = output_channels

    def compute_ragged_image_signature(self, images):
        ragged_spec = tf.RaggedTensorSpec(
            shape=images.shape[1:3] + (self.output_channels,),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )
        return ragged_spec

    def augment_ragged_image(self, image, transformation, **kwargs):
        return self.augment_images(
            image, transformations=transformation, **kwargs
        )

    def augment_images(self, images, transformations=None, **kwargs):
        grayscales = image_utils.rgb_to_grayscale(images)
        if self.output_channels == 1:
            return grayscales
        elif self.output_channels == 3:
            return tf.image.grayscale_to_rgb(grayscales)
        else:
            raise ValueError("Unsupported value for `output_channels`.")

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update({"output_channels": self.output_channels})
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils

# Defining modes for random flipping
HORIZONTAL = "horizontal"
VERTICAL = "vertical"
HORIZONTAL_AND_VERTICAL = "horizontal_and_vertical"


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomFlip(VectorizedBaseRandomLayer):
    """Randomly flips the input images.

    This layer will flip the images horizontally and or vertically based on the
    ``mode`` attribute.

    Args:
        mode (str, optional): The flip mode to use. Supported values:
            ``"horizontal", "vertical", "horizontal_and_vertical"``. Defaults to
            ``"horizontal"``. ``"horizontal"`` is a left-right flip and
            ``"vertical"`` is a top-bottom flip.
        rate (float, optional): The frequency of flipping. ``1.0`` indicates
            that images are always flipped. ``0.0`` indicates no flipping.
            Defaults to ``0.5``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        mode=HORIZONTAL,
        rate=0.5,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if mode == HORIZONTAL:
            self.horizontal = True
            self.vertical = False
        elif mode == VERTICAL:
            self.horizontal = False
            self.vertical = True
        elif mode == HORIZONTAL_AND_VERTICAL:
            self.horizontal = True
            self.vertical = True
        else:
            raise ValueError(
                "RandomFlip layer {name} received an unknown mode="
                "{arg}".format(name=self.name, arg=mode)
            )
        self.mode = mode
        if rate < 0.0 or rate > 1.0:
            raise ValueError(
                f"`rate` should be inside of range [0, 1]. Got rate={rate}"
            )
        self.rate = rate
        self.bounding_box_format = bounding_box_format
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        flip_horizontals = tf.zeros(shape=(batch_size, 1))
        flip_verticals = tf.zeros(shape=(batch_size, 1))
        if self.horizontal:
            flip_horizontals = self._random_generator.random_uniform(
                shape=(batch_size, 1)
            )
        if self.vertical:
            flip_verticals = self._random_generator.random_uniform(
                shape=(batch_size, 1)
            )
        return {
            "flip_horizontals": flip_horizontals,
            "flip_verticals": flip_verticals,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        return self.flip_images(images, transformations)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self, bounding_boxes, transformations=None, raw_images=None, **kwargs
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomFlip()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomFlip(bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="rel_xyxy",
            images=raw_images,
        )
        boxes = bounding_boxes["boxes"]
        flip_horizontals = transformations["flip_horizontals"]
        flip_verticals = transformations["flip_verticals"]
        # broadcast
        flip_horizontals = flip_horizontals[:, tf.newaxis, :]
        flip_verticals = flip_verticals[:, tf.newaxis, :]
        if self.horizontal:
            boxes = tf.where(
                flip_horizontals > (1.0 - self.rate),
                self.flip_boxes_horizontal(boxes),
                boxes,
            )
        if self.vertical:
            boxes = tf.where(
                flip_verticals > (1.0 - self.rate),
                self.flip_boxes_vertical(boxes),
                boxes,
            )
        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="rel_xyxy",
            images=raw_images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="rel_xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=raw_images,
        )
        return bounding_boxes

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations=None, **kwargs
    ):
        return self.flip_images(segmentation_masks, transformations)

    def flip_images(self, images, transformations):
        original_shape = images.shape
        flip_horizontals = transformations["flip_horizontals"]
        flip_verticals = transformations["flip_verticals"]
        # broadcast
        flip_horizontals = flip_horizontals[:, tf.newaxis, tf.newaxis, :]
        flip_verticals = flip_verticals[:, tf.newaxis, tf.newaxis, :]
        if self.horizontal:
            images = tf.where(
                flip_horizontals > (1.0 - self.rate),
                tf.image.flip_left_right(images),
                images,
            )
        if self.vertical:
            images = tf.where(
                flip_verticals > (1.0 - self.rate),
                tf.image.flip_up_down(images),
                images,
            )
        images.set_shape(original_shape)
        return images

    def flip_boxes_horizontal(self, boxes):
        x1, x2, x3, x4 = tf.split(boxes, 4, axis=-1)
        outputs = tf.concat([1 - x3, x2, 1 - x1, x4], axis=-1)
        return outputs

    def flip_boxes_vertical(self, boxes):
        x1, x2, x3, x4 = tf.split(boxes, 4, axis=-1)
        outputs = tf.concat([x1, 1 - x4, x3, 1 - x2], axis=-1)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "mode": self.mode,
                "rate": self.rate,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config
